function Lens(n, index, type, xlist, ylist) {
    if (typeof (xlist) === "undefined") {
        xlist = []
    }
    if (typeof (ylist) === "undefined") {
        ylist = []
    }
    this.type = "lens";
    this.n = n;
    this.index = index;
    this.color = "rgba(0,255,255,0.5)";
    this.xlist = xlist;
    this.ylist = ylist;
    this.transferable = false;
    this.rotatable = false;
    this.anchor = {x: 0, y: 0};
    this.typeOfLen = type;
    this.info = null;
    this.buildfacet = function (xlist, ylist) {
        this.xlist = xlist;
        this.ylist = ylist;
    };
    this.buildfacet2 = function (path) {
        const xlist = [], ylist = [];
        for (let i = 0; i < path.length; i++) {
            xlist[i] = path[i].x;
            ylist[i] = path[i].y;
        }
        this.xlist = xlist;
        this.ylist = ylist;
    };
    this.buildcircle = function (xc, yc, r, begin, end, concave, thick, percise) {
        const xlist = [], ylist = [];
        this.xc = xc;
        this.yc = yc;
        this.r = r;
        this.begin = begin;
        this.end = end;
        this.thick = thick;
        this.concave = concave;
        if (typeof (percise) === "undefined") {
            percise = 1 / r
        }
        if (typeof (begin) === "undefined") {
            begin = 0
        }
        if (typeof (end) === "undefined") {
            end = 2 * Math.PI
        }
        if (typeof (concave) === "undefined") {
            concave = true
        }

        for (let theta = begin; theta < end; theta = theta + percise) {
            xlist.push(r * Math.cos(theta) + xc);
            ylist.push(r * Math.sin(theta) + yc);
        }
        if (concave === false) {
            if (typeof (thick) === "undefined") {
                thick = 5
            }
            let c_angle = (begin + end) / 2, angle_sp = (end - begin) / 2;
            let d = r * (1 - Math.cos(angle_sp)) + thick;
            xlist.push(xc + r * Math.cos(end) + d * Math.cos(c_angle), xc + r * Math.cos(begin) + d * Math.cos(c_angle))
            ylist.push(yc + r * Math.sin(end) + d * Math.sin(c_angle), yc + r * Math.sin(begin) + d * Math.sin(c_angle))
        }
        this.xlist = xlist;
        this.ylist = ylist;
    };
    this.draw = function (context) {
        context.save();
        context.fillStyle = this.color;
        context.beginPath();
        context.moveTo(this.xlist[0], this.ylist[0]);
        for (let i = 1; i < this.xlist.length; i++) {
            context.lineTo(this.xlist[i], this.ylist[i])
        }
        context.fill();
        context.restore();
        context.closePath();
        context.beginPath();
        if (this.rotatable) {
            context.arc(this.anchor.x, this.anchor.y, 4, 0, Math.PI * 2);
        }
        context.fill();
        context.closePath();
    };
    this.transfer = function (x, y) {
        for (let i = 0; i < this.xlist.length; i++) {
            this.xlist[i] = this.xlist[i] + x;
            this.ylist[i] = this.ylist[i] + y;
        }
        if (this.typeOfLen === 1) {
            this.xc += x;
            this.yc += y;
        }
        this.info.replaceChild(this.panel_body(), this.info.firstChild);
    };
    this.rotate = function (theta) {
        const xc = this.anchor.x, yc = this.anchor.y;
        for (let i = 0; i < this.xlist.length; i++) {
            let x = this.xlist[i];
            let y = this.ylist[i];
            this.xlist[i] = (x - xc) * Math.cos(theta) + (y - yc) * Math.sin(theta) + xc;
            this.ylist[i] = -(x - xc) * Math.sin(theta) + (y - yc) * Math.cos(theta) + yc;
        }
        this.info.replaceChild(this.panel_body(), this.info.firstChild);
    };
    this.checkcross = function (light) {
        let i;
        const x0 = light.end.x, y0 = light.end.y;
        const theta = light.theta;
        const xl = this.xlist, yl = this.ylist;
        const n = xl.length;
        let crosspoint_t = [], crosspoint = [];
        const normalset = [];
        let normal;
        const f = function (x1, y1, x2, y2) {
            return (Math.sin(theta) * (x1 - x2) - Math.cos(theta) * (y1 - y2))
        };
//这里注意的是，tan90不存在，所以可能会有bug
        for (i = 0; i < n; i++) {
            //如果有交点
            if (f(xl[i], yl[i], x0, y0) * f(xl[(i + 1) % n], yl[(i + 1) % n], x0, y0) <= 0) {
                const lambda = -f(xl[(i + 1) % n], yl[(i + 1) % n], x0, y0) / f(xl[i], yl[i], xl[(i + 1) % n], yl[(i + 1) % n]);
                const xc = lambda * xl[i] + (1 - lambda) * xl[(i + 1) % n];
                const yc = lambda * yl[i] + (1 - lambda) * yl[(i + 1) % n];
                //在正确的方向
                //还没改完
                // 这里为什么取1个像素点，见mirro说明，也就是两个光线转折点的坐标至少相距1px左右
                // 可以大大增加了计算的稳定性。有时候由于这个因素，有些折射会被忽略。
                if (f(yc, -xc, y0, -x0) > 0.1) {
                    crosspoint_t.push({x: xc, y: yc});
                    normalset.push(Math.atan(-(xl[i] - xl[(i + 1) % n]) / (yl[i] - yl[(i + 1) % n])));
                }
            }
        }
        // 如果存在正确方向的交点
        if (crosspoint_t.length !== 0) {
            //找到最近的点
            crosspoint = crosspoint_t[0];
            normal = normalset[0];
            let d = Math.abs(crosspoint.x - x0) + Math.abs(crosspoint.y - y0);
            for (i = 1; i < crosspoint_t.length; i++) {
                let nd = Math.abs(crosspoint_t[i].x - x0) + Math.abs(crosspoint_t[i].y - y0);
                if (nd < d) {
                    crosspoint = crosspoint_t[i];
                    d = nd;
                    normal = normalset[i];
                }
            }
        }
        //原来的定义是法线具有方向性，如果法线和光线方向夹锐角则是射出，钝角则是射入，
        //现在直接比较折射率属性即可,当然如果允许拼接的时候，要对len赋予新的属性，拼接的边和没有拼接的边是不一样的
        //必须加以区分，
        let outn;
        // 注意，这里没有区分全反射
        // 如果光线的折射率和物质的折射率相同则说明光线在物质内，所以返回的是环境的折射率，作为出射折射率。
        // 但是这必须是保证出射光一定可以出射，如果不可以出射，就
        if (light.n === this.n) {
            outn = 1;
        } else {
            outn = this.n;
        }
        return {cross: crosspoint, normal: normal, n: outn}
    };
    this.show_info = function () {
        return "折射率：" + this.n;
    };
    this.attributes_input = function () {
        let ele = document.createElement("div");
        ele.id = "len" + (this.index + 1);
        ele.classList.add("panel-default");

        let head = document.createElement("div");
        head.id = "len" + (this.index + 1) + "-head";
        head.classList.add("panel-heading");
        head.style.padding = "5px 25px 5px 25px";
        let anchor = document.createElement("a");
        anchor.href = "#len" + (this.index + 1) + "-info";
        anchor.setAttribute("data-toggle", "collapse");
        anchor.setAttribute("data-parent", "#accordion");
        anchor.setAttribute("aria-expanded", "true");
        anchor.setAttribute("onclick", "chevron_toggle(this)");
        let badge = document.createElement("span");
        badge.classList.add("badge");
        badge.innerText = this.index + 1;
        let icon = document.createElement("span");
        icon.classList.add("pull-right", "glyphicon", "glyphicon-chevron-down");
        anchor.appendChild(badge);
        anchor.appendChild(icon);
        head.appendChild(anchor)

        let info = document.createElement("div");
        info.classList.add("panel-collapse", "collapse");
        info.setAttribute("id", "len" + (this.index + 1) + "-info");
        info.setAttribute("aria-expanded", "true");
        let panel_body = this.panel_body();
        info.appendChild(panel_body);
        this.info = info;

        ele.appendChild(head);
        ele.appendChild(info);
        return ele;
    };
    this.panel_body = function () {
        let panel_body = document.createElement("div");
        panel_body.classList.add("panel-body");
        switch (this.typeOfLen) {
            case 0:
                for (let i = 0; i < this.xlist.length * 2 + 1; i++) {
                    let input_group = document.createElement("div");
                    let sp = document.createElement("span");
                    let input = document.createElement("input");
                    input_group.classList.add("input-group", "input-group-sm");
                    sp.classList.add("input-group-addon");
                    input.classList.add("form-control");
                    input.tabIndex = -1;
                    input.type = "number";
                    input.name = "len1," + this.index + ",";
                    input.setAttribute("onchange", "change_values(this);");
                    if (i === this.xlist.length * 2) {
                        sp.innerText = "n";
                        input.value = this.n.toFixed(2);
                        input.step = 0.01;
                        input.name += "n";
                        input.setAttribute("oninput", "if(value<1){value=1};if(value>2){value=2};");
                    } else if (i % 2 === 0) {
                        sp.innerText = "x" + (Math.floor(i / 2) + 1) + ":";
                        console.log(this.xlist, this.xlist[Math.floor(i / 2)])
                        input.value = this.xlist[Math.floor(i / 2)].toFixed(2);
                        input.step = 0.01;
                        input.name += "x" + (Math.floor(i / 2) + 1);
                        input.setAttribute("oninput", "if(value<0){value=0};if(value>800){value=800};");
                    } else {
                        sp.innerText = "y" + (Math.floor(i / 2) + 1) + ":";
                        input.value = this.xlist[Math.floor(i / 2)].toFixed(2);
                        input.step = 0.01;
                        input.name += "y" + (Math.floor(i / 2) + 1);
                        input.setAttribute("oninput", "if(value<0){value=0};if(value>600){value=600};");
                    }
                    input_group.appendChild(sp);
                    input_group.appendChild(input);
                    panel_body.appendChild(input_group);
                }
                break;
            case 1:
                for (let i = 0; i < 6; i++) {
                    let input_group = document.createElement("div");
                    let sp = document.createElement("span");
                    let input = document.createElement("input");
                    input_group.classList.add("input-group", "input-group-sm");
                    sp.classList.add("input-group-addon");
                    input.classList.add("form-control");
                    input.tabIndex = -1;
                    input.type = "number";
                    input.name = "len2," + this.index + ",";
                    input.setAttribute("onchange", "change_values(this);");
                    switch (i) {
                        case 0:
                            sp.innerText = "圆心x:";
                            input.value = this.xc.toFixed(2);
                            input.step = 0.01;
                            input.name += "x";
                            input.setAttribute("oninput", "if(value<0){value=0};if(value>800){value=800};");
                            break;
                        case 1:
                            sp.innerText = "圆心y:";
                            input.value = this.yc.toFixed(2);
                            input.step = 0.01;
                            input.name += "y";
                            input.setAttribute("oninput", "if(value<0){value=0};if(value>600){value=600};");
                            break;
                        case 2:
                            sp.innerText = "半径r:";
                            input.value = this.r.toFixed(2);
                            input.step = 0.01;
                            input.name += "r";
                            input.setAttribute("oninput", "if(value<10){value=10};if(value>600){value=600};");
                            break;
                        case 3:
                            sp.innerText = "起始角:";
                            input.value = (this.begin / Math.PI * 180).toFixed(2);
                            input.step = 0.01;
                            input.name += "begin";
                            input.setAttribute("oninput", "if(value<-360){value=-360};if(value>360){value=360};");
                            break;
                        case 4:
                            sp.innerText = "结束角:";
                            input.value = (this.end / Math.PI * 180).toFixed(2);
                            input.step = 0.01;
                            input.name += "end";
                            input.setAttribute("oninput", "if(value<-360){value=-360};if(value>360){value=360};");
                            break;
                        case 5:
                            sp.innerText = "n:";
                            input.value = this.n.toFixed(2);
                            input.step = 0.01;
                            input.name += "n";
                            input.setAttribute("oninput", "if(value<1){value=1};if(value>2){value=2};");
                            break;
                    }
                    input_group.appendChild(sp);
                    input_group.appendChild(input);
                    panel_body.appendChild(input_group);
                }
                if (!this.concave) {
                    let input_group = document.createElement("div");
                    let sp = document.createElement("span");
                    let input = document.createElement("input");
                    input_group.classList.add("input-group", "input-group-sm");
                    sp.classList.add("input-group-addon");
                    input.classList.add("form-control");
                    input.tabIndex = -1;
                    input.type = "number";
                    input.name = "len2," + this.index + ",";
                    input.setAttribute("onchange", "change_values(this);");

                    sp.innerText = "厚度:";
                    input.value = this.thick.toFixed(0);
                    input.step = 1;
                    input.name += "width";
                    input.setAttribute("oninput", "if(value<1){value=1};if(value>99){value=99};");

                    input_group.appendChild(sp);
                    input_group.appendChild(input);
                    panel_body.appendChild(input_group);
                }
                break;
            case 2:
                // TODO: the panel_body of rect lens
        }
        return panel_body;
    };
}

function Mirros(loc1, loc2, R, index) {
    if (typeof (R) === "undefined") {
        R = 1
    }
    this.type = "mirros";
    this.loc1 = loc1;
    this.loc2 = loc2;
    this.R = R;
    this.index = index;
    this.color = "blue";
    this.transferable = false;
    this.anchor = {x: 0, y: 0};
    this.rotatable = false;
    this.info = null;
    this.draw = function (context) {
        context.save();
        context.beginPath();
        context.strokeStyle = this.color;
        context.lineWidth = 1;
        context.moveTo(loc1.x, loc1.y);
        context.lineTo(loc2.x, loc2.y);
        context.stroke();
        context.restore();
        context.closePath();

        if (this.rotatable) {
            context.beginPath();
            context.arc(this.anchor.x, this.anchor.y, 4, 0, Math.PI * 2);
            context.fill();
            context.closePath();
        }
    };
    this.transfer = function (x, y) {
        this.loc1.x = this.loc1.x + x;
        this.loc1.y = this.loc1.y + y;
        this.loc2.x = this.loc2.x + x;
        this.loc2.y = this.loc2.y + y;
        this.info.replaceChild(this.panel_body(), this.info.firstChild);
    };
    this.rotate = function (theta) {
        const xc = this.anchor.x, yc = this.anchor.y;
        let loc1x = this.loc1.x;
        let loc2x = this.loc2.x;
        this.loc1.x = (loc1x - xc) * Math.cos(theta) + (this.loc1.y - yc) * Math.sin(theta) + xc;
        this.loc1.y = -(loc1x - xc) * Math.sin(theta) + (this.loc1.y - yc) * Math.cos(theta) + yc;
        this.loc2.x = (loc2x - xc) * Math.cos(theta) + (this.loc2.y - yc) * Math.sin(theta) + xc;
        this.loc2.y = -(loc2x - xc) * Math.sin(theta) + (this.loc2.y - yc) * Math.cos(theta) + yc;
        this.info.replaceChild(this.panel_body(), this.info.firstChild);
    };
    this.checkcross = function (light) {
        let normal, cross;
        let xc, yc;
        const theta = light.theta;
        const f = function (p1, p2) {
            return Math.sin(theta) * (p1.x - p2.x) - Math.cos(theta) * (p1.y - p2.y)
        };
        if (f(this.loc1, light.end) * f(this.loc2, light.end) <= 0) {
            const lambda = -f(this.loc2, light.end) / f(this.loc1, this.loc2);
            //也需要在正确的方向上
            xc = lambda * this.loc1.x + (1 - lambda) * this.loc2.x;
            yc = lambda * this.loc1.y + (1 - lambda) * this.loc2.y;
            // 这里一个pix是因为可能由于计算误差，导致为0的项一直大于1，返回始终是改变，一直反射，一直运算，选择分辨率大于一个px
            // 由于这里交点已经和光线方向是共线的，所以距离较远的点是没有影响的。
            // 这里的改动是必须的，否则就会造成无限循环
            if (f({x: yc, y: -xc}, {x: light.end.y, y: -light.end.x}) > 1) {
                cross = {x: xc, y: yc};
                normal = Math.atan((loc2.x - loc1.x) / (loc1.y - loc2.y))
            }
        }
        return {cross: cross, normal: normal}
    };
    this.show_info = function () {
        return "[" + (this.index + 1) + "]  " +
            "Point1:(" + this.loc1.x.toFixed(2) + "," + this.loc1.y.toFixed(2) + ") " +
            "Point2:(" + this.loc2.x.toFixed(2) + "," + this.loc2.y.toFixed(2) + ") " +
            "反射率:" + this.R;
    };
    this.attributes_input = function () {
        let ele = document.createElement("div");
        ele.id = "mirror" + (this.index + 1);
        ele.classList.add("panel-default");

        let head = document.createElement("div");
        head.id = "mirror" + (this.index + 1) + "-head";
        head.classList.add("panel-heading");
        head.style.padding = "5px 25px 5px 25px";
        let anchor = document.createElement("a");
        anchor.href = "#mirror" + (this.index + 1) + "-info";
        anchor.setAttribute("data-toggle", "collapse");
        anchor.setAttribute("data-parent", "#accordion");
        anchor.setAttribute("aria-expanded", "true");
        anchor.setAttribute("onclick", "chevron_toggle(this)");
        let badge = document.createElement("span");
        badge.classList.add("badge");
        badge.innerText = this.index + 1;
        let icon = document.createElement("span");
        icon.classList.add("pull-right", "glyphicon", "glyphicon-chevron-down");
        anchor.appendChild(badge);
        anchor.appendChild(icon);
        head.appendChild(anchor)

        let info = document.createElement("div");
        info.classList.add("panel-collapse", "collapse");
        info.setAttribute("id", "mirror" + (this.index + 1) + "-info");
        info.setAttribute("aria-expanded", "true");
        let panel_body = this.panel_body();
        info.appendChild(panel_body);
        this.info = info;

        ele.appendChild(head);
        ele.appendChild(info);
        return ele;
    };
    this.panel_body = function () {
        let panel_body = document.createElement("div");
        panel_body.classList.add("panel-body");
        for (let i = 0; i < 5; i++) {
            let input_group = document.createElement("div");
            let sp = document.createElement("span");
            let input = document.createElement("input");
            input_group.classList.add("input-group", "input-group-sm");
            sp.classList.add("input-group-addon");
            input.classList.add("form-control");
            input.tabIndex = -1;
            input.type = "number";
            input.name = "mirror," + this.index + ",";
            input.setAttribute("onchange", "change_values(this);");
            switch (i) {
                case 0:
                    sp.innerText = "x1:";
                    input.value = this.loc1.x.toFixed(2);
                    input.step = 0.01;
                    input.name += "x1";
                    input.setAttribute("oninput", "if(value<0){value=0};if(value>800){value=800};");
                    break;
                case 1:
                    sp.innerText = "y1:";
                    input.value = this.loc1.y.toFixed(2);
                    input.step = 0.01;
                    input.name += "y1";
                    input.setAttribute("oninput", "if(value<0){value=0};if(value>600){value=600};");
                    break;
                case 2:
                    sp.innerText = "x2:";
                    input.value = this.loc2.x.toFixed(2);
                    input.step = 0.01;
                    input.name += "x2";
                    input.setAttribute("oninput", "if(value<0){value=0};if(value>800){value=800};");
                    break;
                case 3:
                    sp.innerText = "y2:";
                    input.value = this.loc2.y.toFixed(2);
                    input.step = 0.01;
                    input.name += "y2";
                    input.setAttribute("oninput", "if(value<0){value=0};if(value>600){value=600};");
                    break;
                case 4:
                    sp.innerText = "R:";
                    input.value = this.R;
                    input.max = 1;
                    input.min = 0;
                    input.step = 0.01;
                    input.name += "R";
                    input.setAttribute("oninput", "if(value<0){value=0};if(value>1){value=1};");
            }
            input_group.appendChild(sp);
            input_group.appendChild(input);
            panel_body.appendChild(input_group);
        }
        return panel_body;
    };
}

function change_values(obj) {
    let attr = obj.name.split(",");
    if (attr[0] === "light") {
        if (attr[2] === "x") {
            light_set[attr[1]].begin.x = parseInt(obj.value);
            light_set[attr[1]].init();
        } else if (attr[2] === "y") {
            light_set[attr[1]].begin.y = parseInt(obj.value);
            light_set[attr[1]].init();
        } else if (attr[2] === "theta") {
            light_set[attr[1]].begintheta = parseFloat(obj.value) / 180 * Math.PI;
            light_set[attr[1]].init();
        } else if (attr[2] === "color") {
            let rgba = string2RGB(obj.value);
            light_set[attr[1]].color = obj.value;
            let alpha = light_set[attr[1]].rgba[3];
            rgba.push(alpha);
            light_set[attr[1]].rgba = rgba;
            light_set[attr[1]].init();
        } else if (attr[2] === "alpha") {
            light_set[attr[1]].rgba[3] = parseFloat(obj.value);
            light_set[attr[1]].init();
        } else if (attr[2] === "width") {
            light_set[attr[1]].width = parseInt(obj.value);
            light_set[attr[1]].init();
        }
    } else if (attr[0] === "mirror") {
        if (attr[2] === "x1") {
            mirro_set[attr[1]].loc1.x = parseFloat(obj.value);
        } else if (attr[2] === "y1") {
            mirro_set[attr[1]].loc1.y = parseFloat(obj.value);
        } else if (attr[2] === "x2") {
            mirro_set[attr[1]].loc2.x = parseFloat(obj.value);
        } else if (attr[2] === "y2") {
            mirro_set[attr[1]].loc2.y = parseFloat(obj.value);
        } else if (attr[2] === "R") {
            mirro_set[attr[1]].R = parseFloat(obj.value);
        }
    } else if (attr[0] === "len2") {
        let ele = lens_set[attr[1]];
        if (attr[2] === "x") {
            ele.buildcircle(parseFloat(obj.value), ele.yc, ele.r, ele.begin, ele.end, ele.concave, ele.thick);
        } else if (attr[2] === "y") {
            ele.buildcircle(ele.xc, parseFloat(obj.value), ele.r, ele.begin, ele.end, ele.concave, ele.thick);
        } else if (attr[2] === "r") {
            ele.buildcircle(ele.xc, ele.yc, parseFloat(obj.value), ele.begin, ele.end, ele.concave, ele.thick);
        } else if (attr[2] === "begin") {
            ele.buildcircle(ele.xc, ele.yc, ele.r, parseFloat(obj.value) / 180 * Math.PI, ele.end, ele.concave, ele.thick);
        } else if (attr[2] === "end") {
            ele.buildcircle(ele.xc, ele.yc, ele.r, ele.begin, parseFloat(obj.value) / 180 * Math.PI, ele.concave, ele.thick);
        } else if (attr[2] === "n") {
            ele.n = parseFloat(obj.value);
        } else if (attr[2] === "width") {
            ele.buildcircle(ele.xc, ele.yc, ele.r, ele.begin, ele.end, ele.concave, parseFloat(obj.value));
        }
    }
}
