const canvas = document.getElementById("canvas"),
    context = canvas.getContext("2d"),
    add_light = document.getElementById("add_light"),
    pole_light = document.getElementById("pole"),
    light_color = document.getElementById("light_color"),
    light_alpha = document.getElementById("alpha"),
    light_width = document.getElementById("width"),
    ROb = document.getElementById("refract_only"),
    ro_max = document.getElementById("ro_max"),
    add_mirro = document.getElementById("add_mirro"),
    Rrate_m = document.getElementById("Rrate_m"),
    mirro_size = document.getElementById("mirro_size"),
    mirro_angle = document.getElementById("mirro_angle"),
    transfer = document.getElementById("transfer"),
    rotate = document.getElementById("rotate"),
    paint = document.getElementById("paint"),
    circle = document.getElementById("circle"),
    angle = document.getElementById("angle"),
    paint_n = document.getElementById("paint_n"),
    circle_n = document.getElementById("circle_n"),
    concave = document.getElementById("is_concave"),
    thickness = document.getElementById("thickness"),
    rect = document.getElementById("rect"),
    rect_n = document.getElementById("rect_n"),
    get_info = document.getElementById("info"),
    lights_info = document.getElementById("lights"),
    mirrors_info = document.getElementById("mirrors"),
    lens_info = document.getElementById("lens"),
    clear_all = document.getElementById("clear_all");

let add_pole = 0,
    light_set = [],
    mirro_set = [],
    lens_set = [],
    refract_only = true,
    rotate_ref = Number(),
    paintset = [],
    Cpoint = null,
    angle_info = {begin: 0, end: 2 * Math.PI},
    is_concave = true,
    locp = {paint: null, circle: null, rect: null},
    rect_p = null,
    ele_operate = undefined,
    ele = undefined,
    down = false;

let is_transfer = true,
    is_rotate = false,
    is_paint = false,
    is_circle = false,
    is_rect = false,
    is_delete = false;

// 16to rgb
function string2RGB(color) {
    let r, g, b;
    if (color.length === 4) {
        r = parseInt(color.substring(1, 2) + color.substring(1, 2), 16);
        g = parseInt(color.substring(2, 3) + color.substring(2, 4), 16);
        b = parseInt(color.substring(3) + color.substring(3), 16);
    } else {
        r = parseInt(color.substring(1, 3), 16);
        g = parseInt(color.substring(3, 5), 16);
        b = parseInt(color.substring(5), 16);
    }
    return [r, g, b];
}

function dis(p1, p2) {
    return Math.sqrt((p1.x - p2.x) * (p1.x - p2.x) + (p1.y - p2.y) * (p1.y - p2.y))
}

function rgb2rgba(rgb, a) {
    if (typeof (a) === "undefined") {
        a = 1
    }
    return "rgba(" + rgb[0] + "," + rgb[1] + "," + rgb[2] + "," + a + ")"
}

function modtheta(x, y, r) {
    while (x < 0) {
        x = x + y;
    }
    return (x + r) % (y) - r;
}

function Rreflect(t1, t2, pole, n1, n2) {
    //输入入射角折射角，折射率和偏振方向
    let rE;
    if (pole === 0) {
        rE = (n1 * Math.cos(t1) - n2 * Math.cos(t2)) / (n1 * Math.cos(t1) + n2 * Math.cos(t2));
    } else {
        rE = (n2 * Math.cos(t1) - n1 * Math.cos(t2)) / (n2 * Math.cos(t1) + n1 * Math.cos(t2));
    }
    return rE * rE;
}

function Rrefract(t1, t2, pole, n1, n2) {
    let rE
    if (pole === 0) {
        rE = 2 * n1 * Math.cos(t1) / (n1 * Math.cos(t1) + n2 * Math.cos(t2));
    } else {
        rE = 2 * n1 * Math.cos(t1) / (n2 * Math.cos(t1) + n1 * Math.cos(t2));
    }
    return rE * rE
}


add_light.onclick = function () {
    let rgba = string2RGB(light_color.value);
    let width = parseFloat(light_width.value);
    let alpha = parseFloat(light_alpha.value);
    rgba.push(alpha)
    if (!light_set.length) {
        light_set.push(new Light({x: 100, y: 100}, rgba, width, 800, 0, add_pole, 0));
        lights_info.appendChild(createInput(light_set));
        return;
    }
    for (let pos = 100; pos < canvas.height; pos += 10) {
        let blank = true;
        for (const lightKey in light_set) {
            let light = light_set[lightKey];
            if (light.begin.x === pos && light.begin.y === pos) {
                blank = false;
                break;
            }
        }
        if (blank) {
            light_set.push(new Light({x: pos, y: pos,}, rgba, width, 800, 0, add_pole, light_set.length));
            lights_info.appendChild(createInput(light_set));
            return;
        }
    }
    // 这里是加入光对象，而子光对象只会在分光的时候产生，所以这里全部都加入初始化的光
}

pole_light.onclick = function () {
    if (add_pole === 0) {
        add_pole = 1;
        pole_light.innerHTML = "<b>P波</b>"
    } else {
        add_pole = 0;
        pole_light.innerHTML = "<b>S波</b>"
    }
}

add_mirro.onclick = function () {
    let ms = Number(mirro_size.value);
    let ma = Number(mirro_angle.value) * Math.PI / 180;
    let Rrate = Number(Rrate_m.value);
    if (!mirro_set.length) {
        mirro_set.push(new Mirros({x: 100, y: 100}, {
            x: 100 + ms * Math.cos(ma),
            y: 100 + ms * Math.sin(ma)
        }, Rrate, 0));
        mirrors_info.appendChild(createInput(mirro_set));
        return;
    }
    for (let pos = 100; pos < canvas.height; pos += 10) {
        let blank = true;
        for (const mirrorKey in mirro_set) {
            let mirror = mirro_set[mirrorKey];
            if (mirror.loc1.x === pos && mirror.loc1.y === pos) {
                blank = false;
                break;
            }
        }
        if (blank) {
            mirro_set.push(new Mirros({x: pos, y: pos}, {
                x: pos + ms * Math.cos(ma),
                y: pos + ms * Math.sin(ma)
            }, Rrate, mirro_set.length));
            mirrors_info.appendChild(createInput(mirro_set));
            return;
        }
    }
}

ROb.onclick = function () {
    if (refract_only) {
        refract_only = false;
        ROb.innerHTML = "<b>false<b/>"
    } else {
        refract_only = true;
        ROb.innerHTML = "<b>true<b/>"
    }
}

clear_all.onclick = function () {
    light_set = [];
    mirro_set = [];
    lens_set = [];
    lights_info.innerText = "";
    mirrors_info.innerText = "";
    lens_info.innerText = "";
}

concave.onclick = function () {
    if (is_concave) {
        is_concave = false;
        concave.innerHTML = "<b>凹</b>"
    } else {
        is_concave = true;
        concave.innerHTML = "<b>凸</b>"
    }
}

function changeStates(obj) {
    is_transfer = false;
    is_rotate = false;
    is_paint = false;
    is_circle = false;
    is_rect = false;
    is_delete = false;
    const state = obj.children[0].id;
    eval('is_' + state + '= true');
}

function createInput(set, index = set.length - 1) {
    // let a = document.createElement("a");
    // let li = document.createElement("li");
    // a.href = "#";
    // a.text = set[index].show_info();
    // li.appendChild(a);
    // let panel = document.createElement("div");
    // panel.classList.add("panel-default");
    // panel = ;
    return set[index].attributes_input();
}

/*
旋转操作的步骤是
开启旋转模式，先点击，选定锚点，也就是旋转中心，旋转中心被标出
之后鼠标抬起之后，判断已经可以发生旋转，鼠标变为旋转鼠标，
之后再次点击，这时候判断已经选定锚点，不更新旋转中心。
则以这一点作为参考点，（不一定要在物体内），拖动鼠标，物体绕锚点旋转，
判断终止旋转操作的事件是，如果再次点击旋转中心附近的点，则旋转取消。
于是点击事件有三个判断，是否初始化锚点，是否取消锚点，否则旋转
*/

// 这样操作保证了不存在两个同时开启的状况，关闭的时候不开启，开启的时候关闭



// light2D3的主要是要求在平移的时候，能够更加顺滑。可以快速拖动物体。
canvas.onmousemove = function (event) {
    const loc = {
        x: event.offsetX,
        y: event.offsetY
    };
    if (!is_paint && !is_circle && !is_delete && !is_rect) {
        //旋转就不需要这个操作了
        if (typeof (ele_operate) != "undefined") {
            if (ele_operate.transferable && is_transfer) {
                canvas.style.cursor = "move";
                ele_operate.transfer(loc.x - ele_operate.anchor.x, loc.y - ele_operate.anchor.y);
                ele_operate.anchor = loc;
            }
        }
        if (!down) {
            ele = mouse_on_ele(loc);
        }
        //在上述操作后如果平移过大也会被调整回来，ele始终一样
        if (typeof (ele) != "undefined") {
            get_info.innerHTML = ele.show_info();
            canvas.style.cursor = "pointer";
            if (ele.rotatable && is_rotate && rotate_ref != null) {
                let loc_dot_rotate, theta;
                let p1 = {x: loc.x - ele.anchor.x, y: loc.y - ele.anchor.y}
                let p2 = {x: rotate_ref.x - ele.anchor.x, y: rotate_ref.y - ele.anchor.y}
                loc_dot_rotate = p1.x * p2.x + p1.y * p2.y;
                theta = Math.acos(loc_dot_rotate / dis(loc, ele.anchor) / dis(rotate_ref, ele.anchor));
                if (p1.y * p2.x - p2.y * p1.x > 0) {
                    theta = -theta;
                }
                if (theta < Math.PI + 0.1) {
                    ele.rotate(theta);
                }
                rotate_ref = loc;
            }
            canvas.onmousedown = function () {
                ele_operate = ele;
                down = true;
                if (is_transfer) {
                    ele.transferable = true;
                    ele.anchor = loc;
                }
                // 要防止同时打开两个开关，因为二者共用一个anchor
                if (!ele.rotatable && is_rotate) {
                    ele.rotatable = true;
                    ele.anchor = loc;
                } else {
                    rotate_ref = loc;
                    if (dis(rotate_ref, ele.anchor) < 4) {
                        // 距离小于4个px
                        ele.rotatable = false;
                    }
                }
            }
            canvas.onmouseup = function () {
                ele.transferable = false;
                down = false;
                rotate_ref = null;
                ele_operate = undefined;
            }
        } else if (typeof (ele_operate) == "undefined") {
            //这里是没有操作对象，且鼠标在blank处
            //其实到这里，必然是
            canvas.style.cursor = "auto";
            canvas.onmousedown = function () {
            };
            canvas.onmouseup = function () {
            };
        }
    } else if (is_paint) {
        canvas.style.cursor = "crosshair";

        if (paintset.length > 0) {
            locp.paint = loc;
        }
        // 绘制采用的逻辑是，当第一个点点击时开始绘制，之后每点击一个，加入一条路径，做stroke()
        // 并且末端跟随当前的点，之后当点击的位置距离初始位置很短时，判定为结束绘制，
        // 并且在lens_set中添加一个新的元素，并且清空paintset
        // 折射率从绘制框中读取。

        canvas.onmousedown = function () {
            paintset.push(loc);
            locp.paint = loc;
            if (paintset.length > 2 && dis(loc, paintset[0]) < 5) {
                let p_len = new Lens(Number(paint_n.value), lens_set.length, 0);
                p_len.buildfacet2(paintset);
                lens_set.push(p_len);
                lens_info.appendChild(createInput(lens_set));
                locp.paint = null;
                paintset = [];
            }
        }
    } else if (is_circle) {

        if (Cpoint != null) {
            locp.circle = loc;
        }
        // 圆形绘制方案，鼠标点击，选定圆心，绘制圆心，再次点击选定半径，添加元素，清空记忆
        canvas.onmousedown = function () {
            if (Cpoint == null) {
                Cpoint = loc;
                locp.circle = loc;
            } else {
                let cirlen = new Lens(Number(circle_n.value), lens_set.length, 1);
                cirlen.buildcircle(Cpoint.x, Cpoint.y, dis(Cpoint, loc), angle_info.begin, angle_info.end, is_concave, parseFloat(thickness.value))
                lens_set.push(cirlen);
                lens_info.appendChild(createInput(lens_set));
                locp.circle = null;
                Cpoint = null;
            }
        }
    } else if (is_rect) {
        if (locp.rect != null) {
            rect_p = loc;
        }
        canvas.onmousedown = function () {
            if (locp.rect == null) {
                locp.rect = loc;
            } else {
                let rectlen = new Lens(Number(rect_n.value), lens_set.length, 2);
                rectlen.buildfacet([locp.rect.x, rect_p.x, rect_p.x, locp.rect.x], [locp.rect.y, locp.rect.y, rect_p.y, rect_p.y]);
                lens_set.push(rectlen);
                lens_info.appendChild(createInput(lens_set));
                locp.rect = null;
                rect_p = null;
            }
        }
    } else if (is_delete) {
        ele = mouse_on_ele(loc);
        if (typeof (ele) != "undefined") {
            get_info.innerHTML = ele.show_info();
            canvas.style.cursor = "pointer";
        } else {
            canvas.style.cursor = "auto";
        }
        canvas.onmousedown = function () {
        }
        canvas.onmouseup = function () {
            const ele = mouse_on_ele(loc);
            if (typeof (ele) != "undefined") {
                let index = ele.index;
                switch (ele.type) {
                    case "lens":
                        lens_set.splice(index, 1)
                        lens_info.removeChild(lens_info.children[index]);
                        for (let i = index; i < lens_set.length; i++) {
                            lens_set[i].index = i;
                            lens_info.replaceChild(createInput(lens_set, i), lens_info.children[i]);
                        }
                        break
                    case "mirros":
                        mirro_set.splice(index, 1);
                        mirrors_info.removeChild(mirrors_info.children[index]);
                        for (let i = index; i < mirro_set.length; i++) {
                            mirro_set[i].index = i;
                            mirrors_info.replaceChild(createInput(mirro_set, i), mirrors_info.children[i]);
                        }
                        break
                    case "light":
                        light_set.splice(index, 1);
                        lights_info.removeChild(lights_info.children[index]);
                        for (let i = index; i < light_set.length; i++) {
                            light_set[i].index = i;
                            lights_info.replaceChild(createInput(light_set, i), lights_info.children[i]);
                        }
                        break
                }
            }
        }
    }
}

function mouse_on_ele(loc) {
    for (let i_m = 0; i_m < mirro_set.length; i_m++) {
        let loc1 = mirro_set[i_m].loc1;
        let loc2 = mirro_set[i_m].loc2;
        // 使用一个范围椭圆
        if (dis(loc1, loc) + dis(loc2, loc) < dis(loc1, loc2) + 2) {
            return mirro_set[i_m]
        }
    }
    for (let i_l = 0; i_l < lens_set.length; i_l++) {
        let len = lens_set[i_l];
        context.beginPath();
        context.moveTo(len.xlist[0], len.ylist[0]);
        for (let i = 1; i < len.xlist.length; i++) {
            context.lineTo(len.xlist[i], len.ylist[i])
        }
        if (context.isPointInPath(loc.x, loc.y)) {
            return len;
        }
    }
    for (let i = 0; i < light_set.length; i++) {
        let dist = 30;
        let theta = light_set[i].begintheta;
        let loc1 = light_set[i].begin;
        let loc2 = {x: loc1.x + dist * Math.cos(theta), y: loc1.y + dist * Math.sin(theta)};
        // 如果是旋转的话，需要更大的感应范围
        // 优先操作镜子，其次透镜，最后才是光源
        // if (is_rotate) {
        //     dist = 20;
        // }
        if (dis(loc1, loc) + dis(loc2, loc) < dis(loc1, loc2) + 6) {
            return light_set[i]
        }
    }
}


function light_on() {
    let changed = false;
    // 这是一个自动检测所有光学器件的函数，思路是，找出最近的那个，反射或者透射
    // 反射和透射都要算进去，有反射就一定有透射，关于反射率和透射率的计算和光的偏振有关
    // 所以之后可能要加入相位信息，偏振信息，之类的。
    const d = function (p1, p2) {
        return Math.abs(p1.x - p2.x) + Math.abs(p1.y - p2.y)
    };
    for (let i = 0; i < light_set.length; i++) {
        // 每根子光线才能操作 每根子光线的操作只有两类
        /*
        1，反射，不产生新的子光线
        2，折射，必然产生新的子光线，以折射光为原光线的延伸，反射光为新的子光线
        */
        const child_num = light_set[i].childlight.length;
        const beginset = [], thetaset = [], nset = [];
        const widthset = [], alphaset = [];
        for (let ic = 0; ic < child_num; ic++) {
            // 对每一条子光线操作，
            //console.log(light_set[i].childlight[ic])

            const p0 = light_set[i].childlight[ic].end;
            const theta = light_set[i].childlight[ic].theta;
            let next_ele, next_cross, next_normal, next_action, next_n;
            // 这里一定要用let，否则不会初始化，前面的子光线会都后面的子光线的赋值造成影响。

            for (let i_m = 0; i_m < mirro_set.length; i_m++) {
                let cp = mirro_set[i_m].checkcross(light_set[i].childlight[ic])
                if (typeof (cp.cross) != "undefined") {
                    if (typeof (next_ele) === "undefined") {
                        next_ele = mirro_set[i_m]
                        next_cross = cp.cross;
                        next_action = "mirro"
                        next_normal = cp.normal;
                    } else {
                        if (d(cp.cross, p0) < d(next_cross, p0)) {
                            next_ele = mirro_set[i_m];
                            next_cross = cp.cross;
                            next_normal = cp.normal;
                        }
                    }
                }
            }
            for (let i_l = 0; i_l < lens_set.length; i_l++) {
                let cp = lens_set[i_l].checkcross(light_set[i].childlight[ic])
                if (typeof (cp.normal) != "undefined") {
                    if (typeof (next_ele) === "undefined") {
                        next_ele = lens_set[i_l]
                        next_cross = cp.cross;
                        next_action = "lens"
                        next_n = cp.n;
                        next_normal = cp.normal;
                    } else {
                        if (d(cp.cross, p0) < d(next_cross, p0)) {
                            next_ele = lens_set[i_l];
                            next_cross = cp.cross;
                            next_normal = cp.normal;
                            next_n = cp.n;
                            next_action = "lens"
                        }
                    }
                }
            }
            //console.log(next_cross)
            if (next_action === "mirro") {
                if (light_set[i].childlight[ic].path.length < 40) {
                    light_set[i].childlight[ic].reflect(next_normal, next_cross, next_ele.R);
                    changed = true;
                } else if (light_set[i].childlight[ic].path.length === 40) {
                    light_set[i].childlight[ic].reflect(next_normal, next_cross, 0);
                    changed = true;
                }
            } else if (next_action === "lens") {
                let reflect_info;
                let newn;
                changed = true;
                // n的调整必须在折射之前
                newn = light_set[i].childlight[ic].n;
                reflect_info = light_set[i].childlight[ic].refract(next_normal, next_n, next_cross);
                if (typeof (reflect_info) != "undefined" && reflect_info.ri > 0.0001) {
                    //只有非全反射的时候，而且是新的子光线光强大于0.01时，才能够添加新的子光线，alpha小于0.01已经很难分辨了
                    beginset.push(next_cross);
                    thetaset.push(2 * next_normal - theta + Math.PI);
                    // 反射不改变子光线的折射率
                    nset.push(newn);
                    alphaset.push(reflect_info.ri);
                    widthset.push(reflect_info.rw);
                }
            }
        }
        if (!refract_only) {
            //对于一个有限大的晶体，里面会产生无限多个分支，所以讨论分支的时候必须加上强度
            //可以限定强度低于多少，就不画了，现在只是限制最多分支为10
            if (light_set[i].childlight.length < Number(ro_max.value)) {
                light_set[i].addchilds(beginset, thetaset, nset, alphaset, widthset);
            }
        }

    }
    return changed
    // 如果变化了那么继续作图
}

function draw_all(context) {
    for (let i = 0; i < light_set.length; i++) {
        light_set[i].draw(context);
    }
    for (let i_m = 0; i_m < mirro_set.length; i_m++) {
        mirro_set[i_m].draw(context);
    }
    for (let i_l = 0; i_l < lens_set.length; i_l++) {
        lens_set[i_l].draw(context);
    }
}

function draw_tem(context) {
    if (is_circle && Cpoint != null) {
        let c_angle, angle_sp;
        context.beginPath();
        c_angle = Math.atan((locp.circle.y - Cpoint.y) / (locp.circle.x - Cpoint.x)) + Math.PI * (1 - Math.sign(locp.circle.x - Cpoint.x)) / 2;
        angle_sp = parseFloat(angle.value) / 180 * Math.PI;
        context.arc(Cpoint.x, Cpoint.y, dis(locp.circle, Cpoint), c_angle - angle_sp / 2, c_angle + angle_sp / 2);
        context.stroke();
        context.closePath();
        angle_info.begin = c_angle - angle_sp / 2;
        angle_info.end = c_angle + angle_sp / 2;
    }

    if (is_rect && rect_p != null) {
        context.beginPath();
        context.rect(locp.rect.x, locp.rect.y, rect_p.x - locp.rect.x, rect_p.y - locp.rect.y);
        context.stroke();
        context.closePath();
    }

    if (is_paint && paintset.length > 0) {
        context.beginPath();
        context.moveTo(paintset[0].x, paintset[0].y)
        for (let i = 1; i < paintset.length; i++) {
            context.lineTo(paintset[i].x, paintset[i].y)
        }
        context.lineTo(locp.paint.x, locp.paint.y)
        context.stroke();
        context.closePath();
    }
}

function init() {
    for (let i = 0; i < light_set.length; i++) {
        light_set[i].init();
    }
}

/*
context.clearRect(0,0,canvas.width,canvas.height)
var l1 = new light({x:100,y:100},[255,0,0,1],20,200,0,1);
var m1 = new mirros({x:200,y:50},{x:300,y:150},0.5);
var len1 = new lens(10,[300,350,250,200],[200,250,350,300])
light_set = [l1];
mirro_set = [m1];
lens_set = [len1];
init();
refract_only = false;
while (light_on()){
    light_on();
}
console.log(l1.childlight.length)
draw_all(context);
*/


function main() {
    context.clearRect(0, 0, canvas.width, canvas.height)
    init();
    while (light_on()) {
        light_on();
    }
    draw_all(context);
    draw_tem(context);
    requestAnimationFrame(main);
}

main();

