function Light(begin, rgba, width, length, theta, pole, index) {
    if (typeof (rgba) === "undefined") {
        rgba = [255, 0, 0, 1]
    }
    if (typeof (length) === "undefined") {
        length = 800
    }
    if (typeof (theta) === "undefined") {
        theta = 0
    }
    if (typeof (width) === "undefined") {
        width = 5
    }
    this.type = "light";
    this.length = length;
    this.rgba = rgba;
    this.color = '#' +
        rgba[0].toString(16).padStart(2, '0').toUpperCase() +
        rgba[1].toString(16).padStart(2, '0').toUpperCase() +
        rgba[2].toString(16).padStart(2, '0').toUpperCase();
    this.begin = begin;
    this.begintheta = theta;
    this.beginpole = pole;
    this.index = index;
    // 使用0表示水平偏振1表示垂直偏振
    this.width = width;
    this.childlight = [new ChildLight(this.begin, theta, this.rgba, 1, this.length, this.width, this.beginpole)];
    this.anchor = {x: 0, y: 0};
    this.transferable = false;
    this.rotatable = false;
    this.info = null;
}

function ChildLight(begin, theta, rgba, n, length, width, pole) {
    this.path = [begin];
    this.end = begin;
    this.theta = theta;
    this.n = n;
    this.length = length;
    this.rgb = rgba.slice(0, 3);
    this.Intensityset = [rgba[3]];
    this.endintensity = rgba[3];
    this.widthset = [width];
    this.endwidth = width;
    this.pole = pole;
    //传播过程中不改变偏振的两个大致方向
    //现在更加凸显了，lightbeam对象的重要性
    this.draw = function (context) {
        context.save();
        for (var i = 0; i < this.path.length - 1; i++) {
            context.beginPath();
            context.lineWidth = this.widthset[i];
            context.moveTo(this.path[i].x, this.path[i].y)
            context.lineTo(this.path[i + 1].x, this.path[i + 1].y)
            context.strokeStyle = rgb2rgba(this.rgb, this.Intensityset[i]);
            context.stroke();
            context.closePath();
        }
        context.beginPath();
        context.lineWidth = this.widthset[i];
        context.strokeStyle = rgb2rgba(this.rgb, this.Intensityset[i]);
        context.moveTo(this.path[i].x, this.path[i].y)
        context.lineTo(this.path[i].x + this.length * Math.cos(this.theta), this.path[i].y + this.length * Math.sin(this.theta));
        context.stroke();
        context.restore();
    };
    this.addpath = function (loc) {
        this.path.push(loc);
    };
    this.reflect = function (normal, loc, R) {
        this.addpath(loc);
        this.end = loc;
        this.theta = (2 * normal - this.theta) % (2 * Math.PI) - Math.PI;
        this.widthset.push(this.endwidth);
        this.Intensityset.push(this.endintensity * R)
        this.endintensity = this.endintensity * R;
    };
    this.refract = function (normal, n2, loc) {
        const theta = modtheta(this.theta, 2 * Math.PI, Math.PI);
        let alpha, sin1, sin2;
        normal = modtheta(normal, 2 * Math.PI, Math.PI);
        alpha = modtheta(theta - normal, Math.PI, Math.PI / 2);
        normal = theta - alpha;
        sin1 = Math.sin(alpha);
        sin2 = sin1 * this.n / n2;
        //以折射为主，但是无法折射自动反射。因为不允许有停滞在中间的光
        //先让他自动反射吧。
        if (Math.abs(sin2) < 1) {
            let theta2;
            let reflect_intensity, reflect_width;
            theta2 = Math.asin(sin2);
            reflect_intensity = this.endintensity * Rreflect(alpha, theta2, this.pole, this.n, n2);
            reflect_width = this.endwidth;
            this.endintensity = this.endintensity * Rrefract(alpha, theta2, this.pole, this.n, n2) * n2 / this.n;
            this.Intensityset.push(this.endintensity);
            //分母为0时
            this.endwidth = this.endwidth * Math.cos(theta2) / Math.cos(alpha);
            this.widthset.push(this.endwidth);
            //console.log(normal-theta2)
            this.addpath(loc);
            this.end = loc;
            this.theta = modtheta(normal + theta2, 2 * Math.PI, Math.PI);
            this.n = n2;
            //console.log(this.path)
            /*说明已经进入了晶体，如果第一次而第二次的相同则返回真空的这里只考虑，
            真空和镜子内，不考虑镜子的拼接，如果要考虑不相同的折射率的镜子的拼接的话，
            可以再寻找交点的时候多一个判断。
            */
            return {ri: reflect_intensity, rw: reflect_width};
        } else {
            this.reflect(normal, loc, 1);
        }
    }
}

Light.prototype = {
    addchild: function (begin, theta, n, rgba, width) {
        this.childlight.push(new ChildLight(begin, theta, rgba, n, this.length, width, this.beginpole))
    },
    addchilds: function (beginset, thetaset, nset, alphaset, widthset) {
        for (var i = 0; i < beginset.length; i++) {
            let new_rgba = this.rgba.slice(0, 3);
            new_rgba.push(alphaset[i]);
            this.addchild(beginset[i], thetaset[i], nset[i], new_rgba, widthset[i])
        }
    },
    deletechild: function (i) {
        this.childlight.splice(i, 1)
    },
    draw: function (context) {
        for (var i = 0; i < this.childlight.length; i++) {
            this.childlight[i].draw(context)
        }
        if (this.rotatable) {
            context.beginPath();
            context.arc(this.begin.x, this.begin.y, 4, 0, Math.PI * 2);
            context.fill();
            context.closePath();
        }
    },
    init: function () {
        this.childlight = [new ChildLight(this.begin, this.begintheta, this.rgba, 1, this.length, this.width, this.beginpole)];
    },
    transfer: function (x, y) {
        this.begin = {x: this.begin.x + x, y: this.begin.y + y};
        this.init();
        this.info.replaceChild(this.panel_body(), this.info.firstChild);
        //先改变出发点，然后初始化。
    },
    rotate: function (theta) {
        //this.childlight = [new childlight(this.begin,new_theta,this.rgba,this.length,this.width,this.beginpole)]
        this.begintheta = modtheta(this.begintheta - theta, 2 * Math.PI, Math.PI);
        this.init();
        this.info.replaceChild(this.panel_body(), this.info.firstChild);
    },
    pole_info: function () {
        if (this.beginpole === 1) {
            return "P波"
        } else {
            return "S波"
        }
    },
    show_info: function () {
        return "[" + (this.index + 1) + "]  "
            + "位置：(" + this.begin.x + "," + this.begin.y
            + ") 角度：" + Math.round((this.begintheta) % (2 * Math.PI) / Math.PI * 180)
            + "°  颜色：" + this.rgba
            + " 偏振：" + this.pole_info();
    },
    attributes_input: function () {
        let ele = document.createElement("div");
        ele.id = "light" + (this.index + 1);
        ele.classList.add("panel-default");

        let head = document.createElement("div");
        head.id = "light" + (this.index + 1) + "-head";
        head.classList.add("panel-heading");
        head.style.padding = "5px 25px 5px 25px";
        let anchor = document.createElement("a");
        anchor.href = "#light" + (this.index + 1) + "-info";
        anchor.setAttribute("data-toggle", "collapse");
        anchor.setAttribute("data-parent", "#accordion");
        anchor.setAttribute("aria-expanded", "true");
        anchor.setAttribute("onclick", "chevron_toggle(this)");
        let badge = document.createElement("span");
        badge.classList.add("badge");
        badge.innerText = this.index + 1;
        let icon = document.createElement("span");
        icon.classList.add("pull-right", "glyphicon", "glyphicon-chevron-down");
        anchor.appendChild(badge);
        anchor.appendChild(icon);
        head.appendChild(anchor)

        let info = document.createElement("div");
        info.classList.add("panel-collapse", "collapse");
        info.setAttribute("id", "light" + (this.index + 1) + "-info");
        info.setAttribute("aria-expanded", "true");
        let panel_body = this.panel_body();
        info.appendChild(panel_body);
        this.info = info;

        ele.appendChild(head);
        ele.appendChild(info);
        return ele;
    },
    panel_body: function () {
        let panel_body = document.createElement("div");
        panel_body.classList.add("panel-body");
        for (let i = 0; i < 6; i++) {
            let input_group = document.createElement("div");
            let sp = document.createElement("span");
            let input = document.createElement("input");
            input_group.classList.add("input-group", "input-group-sm");
            sp.classList.add("input-group-addon");
            input.classList.add("form-control");
            input.tabIndex = -1;
            input.type = "number";
            input.name = "light," + this.index + ",";
            input.setAttribute("onchange", "change_values(this);");
            switch (i) {
                case 0:
                    sp.innerText = "x:";
                    input.value = this.begin.x;
                    input.name += "x";
                    input.setAttribute("oninput", "if(value<0){value=0};if(value>800){value=800};");
                    break;
                case 1:
                    sp.innerText = "y:";
                    input.value = this.begin.y;
                    input.name += "y";
                    input.setAttribute("oninput", "if(value<0){value=0};if(value>600){value=600};");
                    break;
                case 2:
                    sp.innerText = "角度:";
                    input.value = Math.round((this.begintheta) % (2 * Math.PI) / Math.PI * 180);
                    input.max = 360;
                    input.min = 0;
                    input.step = 0.01;
                    input.name += "theta";
                    input.setAttribute("oninput", "if(value<0){value=0};if(value>360){value=360};");
                    break;
                case 3:
                    sp.innerText = "光线颜色:";
                    input.type = "color";
                    input.value = this.color;
                    input.style.minWidth = "88%";
                    input.name += "color";
                    break;
                case 4:
                    sp.innerText = "透明度:";
                    input.value = this.rgba[3];
                    input.max = 1;
                    input.min = 0;
                    input.step = 0.1;
                    input.name += "alpha";
                    input.setAttribute("oninput", "if(value<0.1){value=0.1};if(value>1){value=1};");
                    break;
                case 5:
                    sp.innerText = "光线粗细:";
                    input.value = this.width;
                    input.max = 9;
                    input.min = 0.1;
                    input.step = 0.1;
                    input.name += "width";
                    input.setAttribute("oninput", "if(value<0.1){value=0.1};if(value>9){value=9};");
            }
            input_group.appendChild(sp);
            input_group.appendChild(input);
            panel_body.appendChild(input_group);
        }
        return panel_body;
    }
}


function light_beam(lightset) {
    this.lightset = lightset;
    this.add_light = function (lights) {
        for (var i = 0; i < lights.length; i++) {
            this.lightset.push(lights[i])
        }
    };
    this.draw = function () {
        for (var i = 0; i < this.lightset; i++) {
            this.lightset[i].draw();
        }
    };
}